#![allow(clippy::assign_op_pattern)]

use dataplane::core::{Encoder, Decoder, Version};
use std::convert::Infallible;
use std::ops::Deref;
use std::str::FromStr;
use std::collections::BTreeMap;
use std::fmt;

use bytes::Buf;
use bytes::BufMut;

#[derive(Encoder, Decoder, Default, Debug, PartialEq, Clone)]
#[cfg_attr(
    feature = "use_serde",
    derive(serde::Serialize, serde::Deserialize),
    serde(rename_all = "camelCase")
)]
pub struct ManagedConnectorSpec {
    pub name: String,

    pub version: Option<String>,

    #[cfg_attr(feature = "use_serde", serde(rename = "type"))]
    pub type_: String, // syslog, github star, slack

    pub topic: String,

    pub parameters: BTreeMap<String, ManageConnectorParameterValue>,

    pub secrets: BTreeMap<String, SecretString>,
}

#[derive(Encoder, Decoder, Debug, PartialEq, Clone)]
#[cfg_attr(
    feature = "use_serde",
    derive(serde::Serialize),
    serde(rename_all = "camelCase", untagged)
)]
pub enum ManageConnectorParameterValueInner {
    Vec(Vec<String>),
    Map(BTreeMap<String, String>),
    String(String),
}

impl Default for ManageConnectorParameterValueInner {
    fn default() -> Self {
        Self::Vec(Vec::new())
    }
}

#[cfg(feature = "use_serde")]
mod always_string_serialize {
    use super::ManageConnectorParameterValueInner;
    use serde::{Deserializer, Deserialize};
    use serde::de::{self, Visitor, SeqAccess, MapAccess};
    use std::fmt;
    use std::collections::BTreeMap;
    struct ParameterValueVisitor;
    impl<'de> Deserialize<'de> for ManageConnectorParameterValueInner {
        fn deserialize<D>(deserializer: D) -> Result<ManageConnectorParameterValueInner, D::Error>
        where
            D: Deserializer<'de>,
        {
            deserializer.deserialize_any(ParameterValueVisitor)
        }
    }

    impl<'de> Visitor<'de> for ParameterValueVisitor {
        type Value = ManageConnectorParameterValueInner;

        fn expecting(&self, formatter: &mut fmt::Formatter) -> fmt::Result {
            formatter.write_str("string, map or sequence")
        }

        fn visit_none<E>(self) -> Result<Self::Value, E>
        where
            E: de::Error,
        {
            self.visit_str("null")
        }
        fn visit_unit<E>(self) -> Result<Self::Value, E>
        where
            E: de::Error,
        {
            self.visit_str("null")
        }

        fn visit_f64<E>(self, v: f64) -> Result<Self::Value, E>
        where
            E: de::Error,
        {
            self.visit_str(&v.to_string())
        }
        fn visit_bool<E>(self, v: bool) -> Result<Self::Value, E>
        where
            E: de::Error,
        {
            self.visit_str(&v.to_string())
        }

        fn visit_i64<E>(self, v: i64) -> Result<Self::Value, E>
        where
            E: de::Error,
        {
            self.visit_str(&v.to_string())
        }

        fn visit_u64<E>(self, v: u64) -> Result<Self::Value, E>
        where
            E: de::Error,
        {
            self.visit_str(&v.to_string())
        }

        fn visit_str<E>(self, value: &str) -> Result<Self::Value, E>
        where
            E: de::Error,
        {
            Ok(ManageConnectorParameterValueInner::String(
                value.to_string(),
            ))
        }

        fn visit_map<M>(self, mut map: M) -> Result<Self::Value, M::Error>
        where
            M: MapAccess<'de>,
        {
            let mut inner = BTreeMap::new();
            while let Some((key, value)) = map.next_entry::<String, String>()? {
                inner.insert(key.clone(), value.clone());
            }

            Ok(ManageConnectorParameterValueInner::Map(inner))
        }

        fn visit_seq<A>(self, mut seq: A) -> Result<Self::Value, A::Error>
        where
            A: SeqAccess<'de>,
        {
            let mut vec_inner = Vec::new();
            while let Some(param) = seq.next_element::<String>()? {
                vec_inner.push(param);
            }
            Ok(ManageConnectorParameterValueInner::Vec(vec_inner))
        }
    }

    #[test]
    fn deserialize_test() {
        let yaml = r#"
name: kafka-out
parameters:
  param_1: "param_str"
  param_2:
   - item_1
   - item_2
   - 10
   - 10.0
   - true
   - On
   - Off
   - null
  param_3:
    arg1: val1
    arg2: 10
    arg3: -10
    arg4: false
    arg5: 1.0
    arg6: null
    arg7: On
    arg8: Off
  param_4: 10
  param_5: 10.0
  param_6: -10
  param_7: True
  param_8: 0xf1
  param_9: null
  param_10: 12.3015e+05
  param_11: [On, Off]
  param_12: true
secrets: {}
topic: poc1
type: kafka-sink
"#;
        use super::ManagedConnectorSpec;
        let connector_spec: ManagedConnectorSpec =
            serde_yaml::from_str(yaml).expect("Failed to deserialize");
        println!("{:?}", connector_spec);
    }
}

#[derive(Default, Debug, PartialEq, Clone)]
#[cfg_attr(
    feature = "use_serde",
    derive(serde::Serialize, serde::Deserialize),
    serde(transparent)
)]
pub struct ManageConnectorParameterValue(pub ManageConnectorParameterValueInner);

impl From<Vec<String>> for ManageConnectorParameterValue {
    fn from(vec: Vec<String>) -> ManageConnectorParameterValue {
        ManageConnectorParameterValue(ManageConnectorParameterValueInner::Vec(vec))
    }
}
impl From<BTreeMap<String, String>> for ManageConnectorParameterValue {
    fn from(map: BTreeMap<String, String>) -> ManageConnectorParameterValue {
        ManageConnectorParameterValue(ManageConnectorParameterValueInner::Map(map))
    }
}
impl From<String> for ManageConnectorParameterValue {
    fn from(inner: String) -> ManageConnectorParameterValue {
        ManageConnectorParameterValue(ManageConnectorParameterValueInner::String(inner))
    }
}

impl Decoder for ManageConnectorParameterValue {
    fn decode<T: Buf>(&mut self, src: &mut T, version: Version) -> Result<(), std::io::Error> {
        if version >= 8 {
            self.0.decode(src, version)
        } else {
            let mut new_string = String::new();
            new_string.decode(src, version)?;
            self.0 = ManageConnectorParameterValueInner::String(new_string);
            Ok(())
        }
    }
}
impl Encoder for ManageConnectorParameterValue {
    fn write_size(&self, version: Version) -> usize {
        if version >= 8 {
            self.0.write_size(version)
        } else {
            match &self.0 {
                ManageConnectorParameterValueInner::String(ref inner) => inner.write_size(version),
                _ => String::new().write_size(version),
            }
        }
    }

    /// encoding contents for buffer
    fn encode<T: BufMut>(&self, dest: &mut T, version: Version) -> Result<(), std::io::Error> {
        if version >= 8 {
            self.0.encode(dest, version)
        } else {
            match &self.0 {
                ManageConnectorParameterValueInner::String(ref inner) => {
                    inner.encode(dest, version)
                }
                _ => String::new().encode(dest, version),
            }
        }
    }
}

impl ManagedConnectorSpec {
    pub fn version(&self) -> String {
        self.version.clone().unwrap_or_else(|| "latest".to_string())
    }
}

#[derive(Encoder, Decoder, Default, PartialEq, Clone)]
#[cfg_attr(
    feature = "use_serde",
    derive(serde::Serialize, serde::Deserialize),
    serde(rename_all = "camelCase")
)]

/// Wrapper for string that does not reveal its internal
/// content in its display and debug implementation
pub struct SecretString(String);

impl fmt::Debug for SecretString {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str("[REDACTED]")
    }
}

impl fmt::Display for SecretString {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        f.write_str("[REDACTED]")
    }
}

impl FromStr for SecretString {
    type Err = Infallible;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        Ok(Self(s.into()))
    }
}

impl From<String> for SecretString {
    fn from(s: String) -> Self {
        Self(s)
    }
}

impl Deref for SecretString {
    type Target = str;
    fn deref(&self) -> &Self::Target {
        &self.0
    }
}
